#
# rdoc_fill.rb
#
# Copyright 2005 Tero Hasu
#

require 'stringio'

require 'rdoc/markup/simple_markup'
require 'rdoc/markup/simple_markup/inline'

class RDocFormatter
  class MySimpleMarkup < SM::SimpleMarkup
    # list:: An +Enumerable+ of +String+.
    def parse_string_list(list)
      list = list.collect do |string|
        SM::Line.new(string)
      end
      @lines = SM::Lines.new(list)
      return [] if @lines.empty?
      @lines.normalize
      assign_types_to_lines
      group_lines
    end
  end

  attr_reader :lines, :bcount, :ecount

  def initialize(data, op = {})
    if data.respond_to? :to_str
      sio = StringIO.new data
      @lines = []
      for line in sio
        line.rstrip! unless op[:no_rstrip]
        @lines.push line
      end
    else
      @lines = data
    end

    # RDoc does not preserve empty lines at the beginning or end,
    # so we must do it.
    @bcount = 0
    @ecount = 0
    unless op[:no_belp]
      while @lines.first == ''
        @lines.shift
        @bcount += 1
      end
      while @lines.last == ''
        @lines.pop
        @ecount += 1
      end
    end
  end

  def parse
    sm = MySimpleMarkup.new
    @lines = sm.parse_string_list(@lines)
  end

  # q:: A +CodePrint::CodePp+ object.
  def unparse(q)
    pptxt = proc do |line|
      q.space_list(line.txt.split(/\s+/)) do |elem|
        q.text elem
      end
    end

    ppelem = proc do |line|
      case line
      when SM::ListItem
        q.text(line.param)
        case line.type
        when :NOTE
          q.text(":")
        end
        q.text(" ")
        q.align do
          pptxt.call(line)
        end
      when SM::BlankLine
      when SM::Heading
        q.text('=' * line.param)
        q.text(' ')
        q.align do
          pptxt.call(line)
        end
      when SM::Paragraph
        pptxt.call(line)
      when SM::Verbatim
        q.text line.txt
      else
        raise "Unknown SM::Line #{line.inspect}"
      end
      lastelem = line
    end

    first = true
    level = 0
    @bcount.times { q.force_nl }
    for line in @lines
      if line.level > level
        q.indent(1) unless level == 0
        level = line.level
      elsif line.level < level
        q.unindent unless level == 1
        level = line.level
      end

      case line
      when SM::ListStart, SM::ListEnd
        # ignore
      else
        if first
          first = false
        else
          q.force_nl
        end
        ppelem.call(line)
      end

      # Verbatims do have line breaks, and are not followed
      # by BlankLine objects.
      last_verbatim = line.kind_of?(SM::Verbatim)
    end
    q.force_nl unless last_verbatim
    @ecount.times { q.force_nl }
  end
end
